@@ -15,7 +15,7 @@
  */
 
 /*
- * Apache example module.  Provide demonstrations of how modules do things.
+ * Apache example_hooks module.  Provide demonstrations of how modules do things.
  * It is not meant to be used in a production server.  Since it participates
  * in all of the processing phases, it could conceivable interfere with
  * the proper operation of other modules -- particularly the ones related
@@ -26,7 +26,7 @@
  * prefixed with 'x_' instead of 'example_'.
  *
  * To use mod_example_hooks, configure the Apache build with
- * --enable-example and compile.  Set up a <Location> block in your
+ * --enable-example-hooks and compile.  Set up a <Location> block in your
  * configuration file like so:
  *
  * <Location /example>
@@ -328,11 +328,12 @@ static x_cfg *our_cconfig(const conn_rec *c)
 static void example_log_each(apr_pool_t *p, server_rec *s, const char *note)
 {
     if (s != NULL) {
-        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_example: %s", note);
-    } else {
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "mod_example_hooks: %s", note);
+    }
+    else {
         apr_file_t *out = NULL;
         apr_file_open_stderr(&out, p);
-        apr_file_printf(out, "mod_example traced in non-loggable "
+        apr_file_printf(out, "mod_example_hooks traced in non-loggable "
                         "context: %s\n", note);
     }
 }
@@ -703,7 +704,7 @@ static void *x_merge_server_config(apr_pool_t *p, void *server1_conf,
  * declaration near the bottom of this file.)  Note that these may be       *
  * called for situations that don't relate primarily to our function - in   *
  * other words, the fixup handler shouldn't assume that the request has     *
- * to do with "example" stuff.                                              *
+ * to do with "example_hooks" stuff.                                        *
  *                                                                          *
  * With the exception of the content handler, all of our routines will be   *
  * called for each request, unless an earlier handler from another module   *
@@ -741,7 +742,6 @@ static int x_pre_config(apr_pool_t *pconf, apr_pool_t *plog,
      * Log the call and exit.
      */
     trace_startup(ptemp, NULL, NULL, "x_pre_config()");
-
     return OK;
 }
 
@@ -1049,7 +1049,8 @@ static int x_handler(request_rec *r)
                                     r->connection->pool);
     if ((status == APR_SUCCESS) && conn_data) {
         ap_rprintf(r, "  <OL>\n%s  </OL>\n", (char *) conn_data);
-    } else {
+    }
+    else {
         ap_rputs("  <P>No connection-specific callback information was "
                  "retrieved.</P>\n", r);
     }
@@ -1135,7 +1136,6 @@ static int x_pre_connection(conn_rec *c, void *csd)
 static int x_process_connection(conn_rec *c)
 {
     trace_connection(c, "x_process_connection()");
-
     return DECLINED;
 }
 
@@ -1181,7 +1181,6 @@ static int x_post_read_request(request_rec *r)
  */
 static int x_translate_name(request_rec *r)
 {
-
     /*
      * We don't actually *do* anything here, except note the fact that we were
      * called.
