@@ -52,8 +52,8 @@ static int lua_table_set(lua_State *L)
         while ( (badchar = ap_strchr(badchar, '\n')) ) {
             *badchar = ' ';
         }
-        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, t->r, 
-                APLOGNO(02614) "mod_lua: Value for '%s' in table '%s' contains newline!",
+        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, t->r, APLOGNO(02614)
+                      "mod_lua: Value for '%s' in table '%s' contains newline!",
                   key, t->n);
         apr_table_set(t->t, key, replacement);
     }
