@@ -64,10 +64,12 @@ typedef struct {
 
 const char *scgi_sendfile_off = "off";
 const char *scgi_sendfile_on = "X-Sendfile";
+const char *scgi_internal_redirect_off = "off";
+const char *scgi_internal_redirect_on = "Location";
 
 typedef struct {
     const char *sendfile;
-    int internal_redirect;
+    const char *internal_redirect;
 } scgi_config;
 
 
@@ -417,11 +419,29 @@ static int pass_response(request_rec *r, proxy_conn_rec *conn)
         }
     }
 
-    if (conf->internal_redirect && r->status == HTTP_OK) {
-        location = apr_table_get(r->headers_out, "Location");
+    if (r->status == HTTP_OK 
+        && (!conf->internal_redirect /* default === On */
+            || conf->internal_redirect != scgi_internal_redirect_off)) {
+        short err = 1;
+        const char *location_header = conf->internal_redirect ? 
+            conf->internal_redirect : scgi_internal_redirect_on;
+
+        location = apr_table_get(r->err_headers_out, location_header);
+        if (!location) {
+            err = 0;
+            location = apr_table_get(r->headers_out, location_header);
+        }
         if (location && *location == '/') {
             scgi_request_config *req_conf = apr_palloc(r->pool,
                                                        sizeof(*req_conf));
+            if (strcasecmp(location_header, "Location")) {
+                if (err) {
+                    apr_table_unset(r->err_headers_out, location_header);
+                }
+                else {
+                    apr_table_unset(r->headers_out, location_header);
+                }
+            }
             req_conf->location = location;
             req_conf->type = scgi_internal_redirect;
             ap_set_module_config(r->request_config, &proxy_scgi_module,
@@ -431,8 +451,9 @@ static int pass_response(request_rec *r, proxy_conn_rec *conn)
         }
     }
 
-    /* XXX: What could we do with that return code? */
-    (void)ap_pass_brigade(r->output_filters, bb);
+    if (ap_pass_brigade(r->output_filters, bb)) {
+        return AP_FILTER_ERROR;
+    }
 
     return OK;
 }
@@ -564,8 +585,8 @@ static void *create_scgi_config(apr_pool_t *p, char *dummy)
 {
     scgi_config *conf=apr_palloc(p, sizeof(*conf));
 
-    conf->sendfile = NULL;
-    conf->internal_redirect = -1;
+    conf->sendfile = NULL; /* === default (off) */
+    conf->internal_redirect = NULL; /* === default (on) */
 
     return conf;
 }
@@ -576,7 +597,7 @@ static void *merge_scgi_config(apr_pool_t *p, void *base_, void *add_)
     scgi_config *base=base_, *add=add_, *conf=apr_palloc(p, sizeof(*conf));
 
     conf->sendfile = add->sendfile ? add->sendfile: base->sendfile;
-    conf->internal_redirect =   (add->internal_redirect != -1)
+    conf->internal_redirect = add->internal_redirect
                               ? add->internal_redirect
                               : base->internal_redirect;
     return conf;
@@ -601,16 +622,33 @@ static const char *scgi_set_send_file(cmd_parms *cmd, void *mconfig,
 }
 
 
+static const char *scgi_set_internal_redirect(cmd_parms *cmd, void *mconfig,
+                                              const char *arg)
+{
+    scgi_config *conf = mconfig;
+
+    if (!strcasecmp(arg, "Off")) {
+        conf->internal_redirect = scgi_internal_redirect_off;
+    }
+    else if (!strcasecmp(arg, "On")) {
+        conf->internal_redirect = scgi_internal_redirect_on;
+    }
+    else {
+        conf->internal_redirect = arg;
+    }
+    return NULL;
+}
+
+
 static const command_rec scgi_cmds[] =
 {
     AP_INIT_TAKE1("ProxySCGISendfile", scgi_set_send_file, NULL,
                   RSRC_CONF|ACCESS_CONF,
-                  "The name of the X-Sendfile peudo response header or "
+                  "The name of the X-Sendfile pseudo response header or "
                   "On or Off"),
-    AP_INIT_FLAG("ProxySCGIInternalRedirect", ap_set_flag_slot,
-                 (void*)APR_OFFSETOF(scgi_config, internal_redirect),
-                 RSRC_CONF|ACCESS_CONF,
-                 "Off if internal redirect responses should not be accepted"),
+    AP_INIT_TAKE1("ProxySCGIInternalRedirect", scgi_set_internal_redirect, NULL,
+                  RSRC_CONF|ACCESS_CONF,
+                  "The name of the pseudo response header or On or Off"),
     {NULL}
 };
 
