@@ -593,7 +593,7 @@ static const char *set_crypto_passphrase_file(cmd_parms *cmd, void *config,
     while (!(ap_cfg_getline(buffer, sizeof(buffer), file))) {
         args = buffer;
         while (*(arg = ap_getword_conf(cmd->pool, &args)) != '\0') {
-            if (*arg == '#' || *arg == 0) {
+            if (*arg == '#') {
                 break;
             }
             set_crypto_passphrase(cmd, config, arg);
